import re
from collections import namedtuple, OrderedDict
from datetime import timedelta
from typing import Dict

from progress_table import ProgressTable
from tabulate import tabulate
from tqdm import tqdm

COMBINED = "combined"
LOSS = "loss"

TrainerMetric = namedtuple("TrainerMetric", ("epoch", "step", "value"))

metricsRegexString = "(train_metrics|test_metrics|validation_metrics)\\.(.+?)\\.(.*)"
metricsRegex = re.compile(metricsRegexString)


def model_logs_directory(model_id: int) -> str:
    return f"logs_model_{model_id}"


def get_metrics_table():
    table = ProgressTable(
        columns=["epochs", "time"],
        num_decimal_places=4,
        reprint_header_every_n_rows=0,
    )
    # Add feature/metric columns separately so we can customize the width
    table.add_column("feature", width=16)
    table.add_column("metric", width=24)
    table.add_columns(["train", "val", "test"])

    # Horrible monkeypatched hack to make progress bar work in colab.
    # TODO: rework this and the logs / metrics streaming code overall.
    def newprint(*args, **kwargs):
        if "end" in kwargs and kwargs["end"] == "\r":
            kwargs["end"] = ""
            args = ("\r",) + args

        for file in table.files:
            print(*args, **kwargs, file=file)

    table._print = newprint

    return table


def split_sorter(split):
    if split.startswith("train"):
        return 0
    elif split.startswith("vali"):
        return 1
    elif split.startswith("test"):
        return 2
    return 3


def is_llm_model(config: Dict):
    return config.get("model_type") == "llm"


def metric_formatter(val) -> str:
    if isinstance(val, float):
        return f"{val:.4f}"
    return val


def pretty_metric_name(metric: str) -> str:
    return (
        metric.replace("test_metrics.", "test.")
        .replace("train_metrics.", "train.")
        .replace("validation_metrics.", "vali.")
    )


def formatted_time_delta(start: timedelta, end: timedelta) -> str:
    delta = start - end
    days, seconds = delta.days, delta.seconds
    hours = days * 24 + seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    return f"{hours:02}:{minutes:02}:{seconds:02}"


@DeprecationWarning
class MetricsPrintedTable:
    """Maintains a table data structure used for logging metrics.

    ╒════════════╤════════════╤════════╤═════════════╤══════════╤═══════════╕
    │ Survived   │   accuracy │   loss │   precision │   recall │   roc_auc │
    ╞════════════╪════════════╪════════╪═════════════╪══════════╪═══════════╡
    │ train      │     0.7420 │ 0.7351 │      0.7107 │   0.5738 │    0.7659 │
    ├────────────┼────────────┼────────┼─────────────┼──────────┼───────────┤
    │ validation │     0.7079 │ 0.9998 │      0.6061 │   0.6061 │    0.7354 │
    ├────────────┼────────────┼────────┼─────────────┼──────────┼───────────┤
    │ test       │     0.7360 │ 0.7620 │      0.6667 │   0.5538 │    0.7358 │
    ╘════════════╧════════════╧════════╧═════════════╧══════════╧═══════════╛
    ╒════════════╤════════╕
    │ combined   │   loss │
    ╞════════════╪════════╡
    │ train      │ 0.7351 │
    ├────────────┼────────┤
    │ validation │ 0.9998 │
    ├────────────┼────────┤
    │ test       │ 0.7620 │
    ╘════════════╧════════╛
    """

    def __init__(self, metrics: Dict[str, float]):  # noqa
        # Construct a mapping of output_feature -> split -> metric_name
        output_features = {}
        for metric_name in metrics:
            match = metricsRegex.match(metric_name)
            if match:
                split, feature, name = match[1], match[2], match[3]
                if feature not in output_features:
                    output_features[feature] = {}
                if split not in output_features[feature]:
                    output_features[feature][split] = []
                output_features[feature][split].append(name)
        self.output_features = output_features

        self.printed_table = OrderedDict()
        for feature in output_features:
            metric_names = set()
            for split in output_features[feature]:
                metric_names = metric_names.union(set(output_features[feature][split]))
            self.printed_table[feature] = [[feature] + list(metric_names)]
        self.printed_table[COMBINED] = [[COMBINED, LOSS]]
        #
        # # Establish the printed table's order of metrics (used for appending metrics in the right order).
        self.metrics_headers = {}
        for output_feature_name in output_features.keys():
            # [0]: The header is the first row, which contains names of metrics.
            # [1:]: Skip the first column as it's just the name of the output feature, not an actual metric name.
            self.metrics_headers[output_feature_name] = self.printed_table[output_feature_name][0][1:]
        self.metrics_headers[COMBINED] = [LOSS]

        self.add_metrics_to_printed_table(metrics)

    def add_metrics_to_printed_table(self, metrics: Dict[str, float]):
        """Add metrics to tables by the order of the table's metric header."""
        for output_feature_name in self.output_features:
            for split, metric_names in sorted(
                self.output_features[output_feature_name].items(),
                key=lambda x: split_sorter(x[0]),
            ):
                printed_metrics = []
                for metric_name in self.metrics_headers[output_feature_name]:
                    full_metric_name = f"{split}.{output_feature_name}.{metric_name}"
                    if metric_name in metric_names and full_metric_name in metrics:
                        printed_metrics.append(metrics[full_metric_name])
                    else:
                        printed_metrics.append("")
                self.printed_table[output_feature_name].append([split[: split.index("_")]] + printed_metrics)

    def log_info(self):
        tqdm.write("\n")
        for output_feature, table in self.printed_table.items():
            tqdm.write(tabulate(table, headers="firstrow", tablefmt="fancy_grid", floatfmt=".4f"))
        tqdm.write("\n")



from dataclasses import dataclass
from typing import Dict, List, Optional

import pandas as pd
from dataclasses_json import dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GeneratedResponse:
    prompt: str
    response: str
    raw_prompt: Optional[str] = None
    sample_index: Optional[int] = None
    sample_data: Dict[str, any] = None
    context: Optional[str] = None
    model_name: Optional[str] = None
    finish_reason: Optional[str] = None
    generated_tokens: Optional[int] = None
    seed: Optional[int] = None
    prefill: Optional[List[Dict]] = None
    tokens: Optional[List[Dict]] = None
    finish_reason: Optional[str] = None
    best_of_sequences: Optional[List[Dict]] = None

    @staticmethod
    def to_responses(df: pd.DataFrame) -> List["GeneratedResponse"]:
        return [GeneratedResponse(**row.to_dict()) for _, row in df.iterrows()]

    @staticmethod
    def to_pandas(responses: List["GeneratedResponse"]) -> pd.DataFrame:
        return pd.DataFrame([response.to_dict() for response in responses])

    def __str__(self):
        return f"GeneratedResponse\n\tprompt: {self.prompt}\n\tresponse: {self.response}\n"



from typing import Dict, List, Optional

from tabulate import tabulate


def print_events(events: Dict, print_header: Optional[bool] = True, existing_events: Optional[List] = []):
    compute_events = [(event["eventType"], event["timestamp"]) for event in events.get("ComputeEvents", [])]
    unique_new_events = [event for event in compute_events if event not in existing_events]
    if len(unique_new_events) > 0:
        if print_header:
            print(f"Resource Name: {events.get('name', '')}\n")
            headers = ["Event Type", "Timestamp", "Resource Active Time"]
            table = tabulate(unique_new_events, headers=headers, tablefmt="simple")
        else:
            table = tabulate(unique_new_events, tablefmt="simple")
        print(table)
    print(f"Resource Active Time (seconds): {events.get('resourceActiveTime', None)}", end="\r", flush=True)


def get_quantization_parameters(quantization_kwargs: Dict):
    """Validates and extracts parameters for quantization.

    `quantization_args` should either be None or be of the following format
    {
        "quantize": "some_quantization method",
        # "parameters" key is optional
        "parameters": [
            # we only allow exactly the following list of dictionaries or none at all.
            {"name": "GPTQ_BITS", "value": "4"},
            {"name": "GPTQ_GROUPSIZE", "value": "128"},
        ]
    }
    """
    environment_variables = None
    quantize = None
    if quantization_kwargs is not None:
        if not isinstance(quantization_kwargs, dict):
            raise ValueError("'quantization_kwargs' must be a dictionary.")
        if not all(key in ["quantize", "parameters"] for key in quantization_kwargs):
            raise ValueError("Only 'quantize' or 'parameters' are allowed as keys in 'quantization_kwargs'.")
        if "quantize" not in quantization_kwargs:
            raise ValueError("'quantize' must exist as a key in 'quantization_kwargs'.")
        quantize = quantization_kwargs["quantize"]
        if not isinstance(quantize, str):
            raise ValueError("'quantize' must be a string.")
        if "parameters" in quantization_kwargs:
            parameters = quantization_kwargs["parameters"]
            if not (isinstance(parameters, list) and parameters and len(parameters) == 2):
                raise ValueError(
                    "'parameters' must be provided as a non-empty list of the following format: "
                    "[{'name': 'GPTQ_BITS', 'value': 'some_value'}, "
                    "{'name': 'GPTQ_GROUPSIZE', 'value': 'some_other_value'}].",
                )
            for param in parameters:
                if not isinstance(param, dict):
                    raise ValueError("Each parameter must be a dictionary.")
                if len(param) != 2:
                    raise ValueError("You must provide two keys: 'name' and 'value'.")
                if "name" not in param or "value" not in param:
                    raise ValueError("You must provide two keys: 'name' and 'value'.")
                if param["name"] not in ["GPTQ_BITS", "GPTQ_GROUPSIZE"]:
                    raise ValueError(
                        "Only 'GPTQ_GROUPSIZE' and 'GPTQ_BITS' quantization parameters are allowed under 'name'.",
                    )
                param["value"] = str(param["value"])
                if not isinstance(param["value"], str):
                    raise ValueError(f"The value of {param['name']} must be a string.")
            environment_variables = parameters
    return quantize, environment_variables



import enum


class VisualizeType(enum.Enum):
    CONFUSION_MATRIX = "confusion_matrix"
    ROC_CURVES_FROM_TEST_STATISTICS = "roc_curves_from_test_statistics"
    FREQUENCY_VS_F1 = "frequency_vs_f1"



import enum

from predibase.resource.user import User


class PermissionNamespace(enum.Enum):
    DATASETS = "datasets"
    CONNECTIONS = "connections"

    ENGINES = "engines"
    MODELS = "models"


class PermissionRelation(enum.Enum):
    # General
    READ = "read"
    WRITE = "write"
    OWN = "own"

    # Engines
    USE = "read"


class PermissionSubjectType(enum.Enum):
    USER = "user"
    ROLE = "role"
    RESERVED_ROLE = "reserved role"


class PermissionRequest:
    def __init__(
        self,
        namespace: PermissionNamespace,
        data_id: int,
        subject_id: str = None,
        subject_name: str = None,
        subject_type: str = PermissionSubjectType,
        relation: str = PermissionRelation,
        is_revoke: bool = False,
    ):
        self.namespace = namespace
        self.data_id = data_id
        self.subject_id = subject_id
        self.subject_name = subject_name
        self.subject_type = subject_type
        self.relation = relation
        self.is_revoke = is_revoke

    def to_dict(self):
        return {
            "namespace": self.namespace.value,
            "dataID": self.data_id,
            "subjectID": self.subject_id,
            "subjectName": self.subject_name,
            "subjectType": self.subject_type.value,
            "relation": self.relation.value,
            "isRevoke": self.is_revoke,
        }


class PermissionMixin:
    def get_current_user(self) -> User:
        resp = self.session.get_json("/users/current")
        return User.from_dict({"session": self.session, **resp})

    def modify_resource_permission(
        self,
        namespace: PermissionNamespace,
        data_id: int,
        subject_id: str,
        subject_type: PermissionSubjectType,
        relation: PermissionRelation,
        is_revoke: bool = False,
    ):
        request = PermissionRequest(namespace, data_id, subject_id, None, subject_type, relation, is_revoke)
        self.session._post("/permissions/", json=request.to_dict())
        return True

    def check_requester_own(self, namespace: PermissionNamespace, data_id: int) -> bool:
        endpoint = f"/permissions/own?namespace={namespace.value}&dataID={data_id}"
        resp = self.session._get(endpoint)
        return resp

    def get_resource_permissions(self, namespace: PermissionNamespace, data_id: int):
        endpoint = f"/permissions?namespace={namespace.value}&dataID={data_id}"
        resp = self.session.get_json(endpoint)
        return resp



from dataclasses import dataclass, field
from typing import Optional

import pandas as pd
from dataclasses_json import config, dataclass_json, LetterCase

from predibase.pql.api import Session
from predibase.predictor import AsyncPredictor, Predictor
from predibase.util import spinner


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Deployment:
    session: Session
    """['deployment_name', 'deployment_version', 'engine_name', 'model_name', 'model_version', 'deployment_url',
    'comment', 'error_text'],"""
    _name: int = field(metadata=config(field_name="name"))
    _deployment_version: int = field(metadata=config(field_name="deploymentVersion"))
    _engine_name: str = field(metadata=config(field_name="engineName"))
    _model_name: Optional[int] = field(metadata=config(field_name="modelName"))
    _model_version: Optional[int] = field(metadata=config(field_name="modelVersion"))
    _deployment_url: Optional[str] = field(metadata=config(field_name="deploymentUrl"))
    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    _error_text: Optional[str] = field(metadata=config(field_name="errorText"))

    @spinner(name="Predict")
    def predict(self, input_df: pd.DataFrame, stream=False) -> pd.DataFrame:
        """Make predictions for each row of the pandas dataframe.

        :param input_df pandas DataFrame: Required input dataframe.
        :param bool stream: Optional flag to stream results via gRPC protocol (default False).
        :return: pandas dataframe with predictions.
        """
        if not self.session.tenant:
            raise Exception("tenant must be set on session before calling predict")
        predictor = Predictor(self.session, deployment_name=self._name, deployment_version=self._deployment_version)
        results = predictor.predict(input_df, stream)
        if results:
            return results.to_pandas()
        return pd.DataFrame()

    @spinner(name="Async Predict")
    async def async_predict(self, input_df: pd.DataFrame, stream=False) -> pd.DataFrame:
        """Make predictions for each row of the pandas dataframe.

        :param input_df pandas DataFrame: Required input dataframe.
        :param bool stream: Optional flag to stream results via gRPC protocol (default False).
        :return: pandas dataframe with predictions.
        """
        if not self.session.tenant:
            raise Exception("tenant must be set on session before calling predict")
        predictor = AsyncPredictor(
            self.session,
            deployment_name=self._name,
            deployment_version=self._deployment_version,
        )
        results = await predictor.predict(input_df, stream)
        if results:
            return results.to_pandas()
        return pd.DataFrame()

    def __repr__(self):
        return (
            f"Deployment(name={self.name}, deployment_version={self.deployment_version}, "
            f"engine_name={self.engine_name}, model_name={self.model_name}, "
            f"model_version={self.model_version}, deployment_url={self.deployment_url}, "
            f"comment='{self.comment}', error_text='{self.error_text}')"
        )

    @property
    def name(self):
        """Get name."""
        return self._name

    @property
    def deployment_version(self):
        """Get deployment version."""
        return self._deployment_version

    @property
    def engine_name(self):
        """Get engine name."""
        return self._engine_name

    @property
    def model_name(self):
        """Get model name."""
        return self._model_name

    @property
    def model_version(self):
        """Get model version."""
        return self._model_version

    @property
    def deployment_url(self):
        """Get deployment url."""
        return self._deployment_url

    @property
    def comment(self):
        return self._comment

    @property
    def error_text(self):
        return self._error_text

    def __dict__(self):
        return {
            "name": self.name,
            "deploymentVersion": self.deployment_version,
            "engineName": self.engine_name,
            "modelName": self.model_name,
            "modelVersion": self.model_version,
            "deploymentUrl": self.deployment_url,
            "comment": self.comment,
            "errorText": self.error_text,
        }



from typing import List, Optional

import pandas as pd
import typer
from rich import box
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from predibase.cli_commands.utils import df_to_table, get_client, get_console

app = typer.Typer(no_args_is_help=True)


@app.command(help="Query a Large Language Model (LLM)")
def llm(
    data: List[str] = typer.Argument(
        default=None,
        help="A single string or a space-separated list of <field_name>=<field_value> pairs",
    ),
    deployment_name: str = typer.Option(
        None,
        "--deployment-name",
        "-m",
        prompt="Deployment to prompt",
        prompt_required=True,
        help="The name of the deployment to prompt",
    ),
    adapter_name: Optional[str] = typer.Option(
        None,
        "--adapter-name",
        "-a",
        prompt="Optional fine-tuned adapter model",
        prompt_required=False,
        help="Reference to a fine-tuned adapter model with the same base model type",
    ),
    dataset_name: Optional[str] = typer.Option(
        None,
        "--dataset-name",
        "-d",
        prompt_required=False,
        help="Optional dataset for batch inference",
    ),
    limit: Optional[int] = typer.Option(None, "--limit", "-l", prompt_required=False, help="Optional results limit"),
):
    if len(data) == 1:
        data = data[0]
    else:
        dd = {}
        for s in data:
            k, v = s.split("=", 1)
            dd[k] = v
        data = dd

    client = get_client()
    dep = client.LLM(f"pb://deployments/{deployment_name}")

    if adapter_name:
        if adapter_name.startswith("hf://"):
            model = client.LLM(adapter_name)
        elif adapter_name.startswith("pb://models/"):
            segments = adapter_name[len("pb://models/") :].split("/")
            version = segments[1] if len(segments) > 1 else None
            model = client.get_model(name=segments[0], version=version)
        else:
            raise ValueError(
                f"unrecognized model name {adapter_name} - "
                f"expected either a Hugging Face reference (e.g., 'hf://meta-llama/Llama-2-7b-hf') "
                f"or a Predibase model reference (e.g., pb://models/<model-repo> or "
                f"pb://models/<model-repo>/<model-version>",
            )

        dep = dep.with_adapter(model)

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Querying LLM...", total=None)
        response = dep.prompt(data)
        df = pd.DataFrame([response.to_dict()])

    table = Table(show_header=True, header_style="bold magenta")

    # Modify the table instance to have the data from the DataFrame
    table = df_to_table(df, table)

    # Update the style of the table
    table.row_styles = ["none", "dim"]
    table.box = box.SIMPLE_HEAD

    get_console().print(table)


if __name__ == "__main__":
    app()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional, Dict, Tuple, Any, Union

import yaml
from pydantic import BaseModel, Field, validator, PrivateAttr
from tabulate import tabulate

from predibase._errors import PredibaseError
from predibase.resource.llm.prompt import PromptTemplate
from predibase.resources.datasets import Datasets
from predibase.resources.model import PretrainedHuggingFaceLLM, FinetunedLLMAdapter
from predibase.util import log_info, get_url

if TYPE_CHECKING:
    from predibase import Predibase
    from predibase.resource.model import ModelRepo

_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_TEMPLATE_DIR = os.path.join(_PATH_HERE, "../resource/llm/templates")
_CONFIG_FILENAME = "config.yaml"  # Name of config file for loading templates.


class Adapters:
    def __init__(self, client: Predibase):
        self._client = client
        self._session = client._session  # Directly using the session in the short term as we transition to v2.

    def create(
        self,
        config: Union[FinetuneConfig, Dict[str, Any]],
        *,
        description: str | None = None,
        repo: str | ModelRepo | None = None,
        wait: bool = True,
        show_tensorboard: bool = True,
    ) -> FinetunedLLMAdapter:
        """
        Triggers training of a fine-tuned adapter.

        :param config: A config object for the training run. Determines, among other parameters, the base LLM to
            fine-tune and the dataset to fine-tune on.
        :param description: (Optional) A human-readable description for this model.
        :param repo: (Optional) The repository to associate this model with. If not provided, one will be automatically
            created.
        :param wait: If True, the `create` call will block until training has completed.
        :param show_tensorboard: If True, training metrics will be streamed and displayed via a local Tensorboard
            instance. Requires `wait` to be True.
        :return: A FinetunedLLMAdapter object representing the (possible still training) adapter.
        """

        if show_tensorboard and not wait:
            raise RuntimeError("`show_tensorboard` is a blocking option and thus requires `wait` to be True.")

        if isinstance(config, dict):
            config = FinetuneConfig(**config)

        if repo is None:
            # If no repo is specified, automatically construct the repo name from the dataset and model name.
            dataset_name = config.dataset
            if "/" in dataset_name:
                _, dataset_name = dataset_name.split("/")

            model_name = config.base_model.name
            if "/" in model_name:
                _, model_name = model_name.split("/")

            repo = f"{model_name}-{dataset_name}"

        if isinstance(repo, str):
            if "/" in repo:
                repo = repo.replace("/", "-")
            repo = self._client.repos.create(repo, exists_ok=True)

        # TODO: this could move to server-side to reduce round trips.
        dataset = Datasets(self._session).get(config.dataset)

        train_req = {
            "config": config.render(),
            "datasetID": dataset.id,
            "repoID": repo.id,
            "description": description,
        }

        resp = self._session.post_json("/models/train", train_req)
        model_id = resp["model"]["id"]

        # Output link to UI model page
        endpoint = f"models/version/{model_id}"
        url = get_url(self._session, endpoint)
        log_info(f"Check Status of Model Training Here: [link={url}]{url}")

        adapter = FinetunedLLMAdapter(self._client, repo.name, model_id)
        if wait:
            return adapter.wait_ready(show_tensorboard=show_tensorboard)
        return adapter

    @staticmethod
    def FinetuneConfig(*args, **kwargs) -> FinetuneConfig:
        return FinetuneConfig(*args, **kwargs)

    def get(self, resource_id: str) -> FinetunedLLMAdapter:
        """
        Retrieve a FinetunedLLMAdapter given its name or URI.
        :param resource_id: A string of the form [pb://adapters/]<repo_name>/<version> identifying the adapter to fetch.
        :return: A FinetunedLLMAdapter object.
        """
        if resource_id.startswith("pb://adapters/"):
            resource_id = resource_id[len("pb://adapters/") :]

        tokens = resource_id.split("/")
        if len(tokens) < 2:
            raise PredibaseError(f"expected a URI of the form [pb://adapters/]<repo_name>/<version>")

        resp = self._session.get_json(f"/models/version/name/{tokens[0]}?version={tokens[1]}")
        if not resp.get("llmBaseModelName", ""):
            raise PredibaseError(f"resource ID {resource_id} does not point to a fine-tuned LLM adapter")

        return FinetunedLLMAdapter(self._client, tokens[0], resp["id"])


# TODO: remove `arbitrary_types_allowed` after moving more classes to Pydantic models
class FinetuneConfig(BaseModel, validate_assignment=True, arbitrary_types_allowed=True):
    base_model: Union[str, PretrainedHuggingFaceLLM] = Field(
        description="The model to fine-tune, expressed either as a string ("
        "e.g., `meta-llama/Llama-2-7b-hf`) or as a "
        "PretrainedHuggingFaceLLM instance."
    )
    # TODO: improve passing dataset by name once we get rid of connection namespacing
    dataset: str = Field(
        description="The dataset to fine-tune on. Accepts a string of the form <connection>/<name> referencing data "
        "uploaded to Predibase."
    )
    prompt_template: Union[str, PromptTemplate] = Field(
        description="The text template used to format each training example. Column data "
        "will be interpolated via `{column_name}` placeholders. A "
        "PromptTemplate object can be used instead in cases where curly "
        "braces are difficult to escape, such as when working with JSON-like "
        "prompts."
    )
    target: str = Field(
        description="The name of the dataset column containing outputs that the model should learn to " "generate."
    )
    finetune_template: FinetuneTemplate = Field(
        default_factory=lambda: FinetuneTemplateCollection().default,
        description="(Optional) A base configuration template that this " "config should take default values from.",
    )
    epochs: Optional[int] = Field(
        default=None,
        description="(Optional) The number of epochs to train for. If not set, "
        "a default value will be used. At most one of this parameter and "
        "`train_steps` may be set.",
    )
    train_steps: Optional[int] = Field(
        default=None,
        description="(Optional) The number of steps to train for. If not set, "
        "a default value will be used. At most one of this parameter and "
        "`epochs` may be set.",
    )
    learning_rate: Optional[float] = Field(
        default=None,
        description="(Optional) The learning rate of the model. If not set, " "a default value will be used.",
    )

    @classmethod
    @validator("train_steps")
    def check_train_steps(cls, v, values):
        if values.get("epochs") is not None:
            raise ValueError("only one of `epochs` and `train_steps` may be set")

    def __init__(self, **data):
        super().__init__(**data)

        if isinstance(self.base_model, str):
            self.base_model = PretrainedHuggingFaceLLM(self.base_model)

    def render(self) -> Dict:
        prompt_template = self.prompt_template
        if isinstance(prompt_template, PromptTemplate):
            prompt_template = prompt_template.render()

        config = self.finetune_template.to_config(self.base_model, prompt_template, self.target)

        # HACK(Arnav): Temporary hack to inject the right target_modules into the config for mixtral and
        # mixtral instruct. This should be removed once the Mixtral models have a default PEFT mapping for LoRA
        # target_modules. See MLX-1680: https://linear.app/predibase/issue/MLX-1680/remove-peftlora-mixtral-hack-in
        # -the-predibase-sdk # noqa
        if (
            config.get("base_model", "") in {"mistralai/Mixtral-8x7B-v0.1", "mistralai/Mixtral-8x7B-Instruct-v0.1"}
            and "adapter" in config
            and config.get("adapter", {}).get("target_modules", None) is None
        ):
            config["adapter"]["target_modules"] = ["q_proj", "v_proj"]

        # Apply first-class training parameters.
        if self.train_steps is not None:
            config["trainer"]["train_steps"] = self.train_steps
        if self.epochs is not None:
            config["trainer"]["epochs"] = self.epochs
        if self.learning_rate is not None:
            config["trainer"]["learning_rate"] = self.learning_rate

        return config


class FinetuneTemplateCollection(dict):
    def __init__(self):
        super().__init__()

        for filename in os.listdir(_TEMPLATE_DIR):
            if not filename.endswith(".yaml"):
                continue

            if filename == _CONFIG_FILENAME:
                with open(os.path.join(_TEMPLATE_DIR, _CONFIG_FILENAME)) as config_file:
                    self._config = yaml.safe_load(config_file)
            else:
                tmpl = FinetuneTemplate(filename=os.path.join(_TEMPLATE_DIR, filename))
                self[tmpl.name] = tmpl

    @property
    def default(self) -> FinetuneTemplate:
        default_name = self._config.get("default", "")
        tmpl = self.get(default_name, None)
        if tmpl is None:
            raise RuntimeError(f"the default template '{default_name}' was not found")

        return tmpl

    def compare(self):
        print(self)

    def __str__(self):
        def make_row(key) -> Tuple[str, str, str]:
            tmpl = self[key]
            default_indicator = "------>" if tmpl.name == self._config["default"] else ""
            return default_indicator, key, tmpl.description

        return tabulate(
            (make_row(k) for k in sorted(self.keys())),
            tablefmt="simple_grid",
            headers=["Default", "Name", "Description"],
            colalign=["center"],
        )


class FinetuneTemplate(BaseModel):
    filename: str
    _meta: Optional[dict] = PrivateAttr(default=None)
    _raw_template_str: Optional[str] = PrivateAttr(default=None)

    @property
    def name(self) -> str:
        if self._meta is None:
            self._load()
        return self._meta["name"]

    @property
    def description(self) -> str:
        if self._meta is None:
            self._load()
        return self._meta["description"]

    def to_config(
        self,
        model: str | PretrainedHuggingFaceLLM,
        prompt_template: str = "__PROMPT_TEMPLATE_PLACEHOLDER__",
        target: str = "__TARGET_PLACEHOLDER__",
    ) -> dict:
        if self._raw_template_str is None:
            self._load()

        if isinstance(model, PretrainedHuggingFaceLLM):
            model = model.name

        cfg = yaml.safe_load(
            self._raw_template_str.format(
                base_model=model,
                prompt_template=prompt_template,
                target=target,
            ),
        )
        del cfg["template_meta"]
        return cfg

    def _load(self):
        with open(os.path.join(_TEMPLATE_DIR, self.filename)) as f:
            self._raw_template_str = f.read()

        self._meta = yaml.safe_load(self._raw_template_str)["template_meta"]



import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union

from dataclasses_json import config, dataclass_json, LetterCase

from predibase.pql.api import Session


class DefaultEngineType(Enum):
    QUERY = "query"
    TRAIN = "train"


def get_replicas_str(node_dict: Dict) -> Union[str, None]:
    replicas = node_dict["replicas"]
    if replicas == node_dict["maxReplicas"] == 0:
        return None
    if node_dict["minReplicas"] != node_dict["maxReplicas"]:
        replicas = f"{replicas}x ({node_dict['minReplicas']} -> {node_dict['maxReplicas']})"
    return f"{replicas}x"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class EngineTemplate:
    id: int
    name: str
    description: str
    service_type: str
    head_node_group: dict = field(metadata=config(field_name="headNodeGroup"))
    worker_node_groups: List[dict] = field(metadata=config(field_name="workerNodeGroups"))
    adaptive: bool = False
    disabled: bool = False
    suggested_workers: Optional[int] = field(default=0, metadata=config(field_name="suggestedWorkers"))
    max_workers: Optional[int] = field(default=0, metadata=config(field_name="maxWorkers"))

    def __str__(self):
        head = "Head: None"
        head_replicas = self.head_node_group["replicas"]

        if head_replicas != 0:
            cloud = next(iter(self.head_node_group["nodeType"]["clouds"]))
            cloud_engine_node_config = self.head_node_group["nodeType"]["clouds"][cloud]
            head_label = cloud_engine_node_config.get("publicLabel", cloud_engine_node_config["label"])
            head = f"Head: {get_replicas_str(self.head_node_group)} {head_label}"

        worker_title = "Workers:"
        workers = "\n".join(
            [
                f"- {get_replicas_str(worker)} {worker['nodeType']['name']}"
                for worker in self.worker_node_groups
                if get_replicas_str(worker) is not None
            ],
        )
        if workers.strip() == "":
            return head + "\nWorkers: None"
        return head + "\n" + worker_title + "\n" + workers


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Engine:
    session: Session
    _id: int = field(metadata=config(field_name="id"))
    _type: str = field(metadata=config(field_name="type"))
    _name: str = field(metadata=config(field_name="name"))
    _auto_suspend_seconds: Optional[int] = field(metadata=config(field_name="autoSuspendSeconds"))
    _template_id: int = field(metadata=config(field_name="templateID"))
    _environment_id: int = field(metadata=config(field_name="environmentID"))
    _service_type: str = field(metadata=config(field_name="serviceType"))
    _auto_resume: Optional[bool] = field(metadata=config(field_name="autoResume"), default=False)

    # _user: dict = field(metadata=config(field_name="user"))

    def __repr__(self):
        return (
            f"Engine(id={self.id}, name={self.name}, status={self.status}, type={self.type}, "
            f"service_type={self.service_type}, template_id={self.template_id}, environment_id={self.environment_id}, "
            f"auto_suspend_seconds={self.auto_suspend_seconds}, auto_resume={self.auto_resume})"
        )

    @property
    def id(self):
        return self._id

    @property
    def type(self):
        return self._type

    @property
    def service_type(self):
        return self._service_type

    @property
    def name(self):
        return self._name

    # @property
    # def author(self):
    #     if
    #     return self._user["username"]

    def set_name(self, value):
        self._name = value
        self._update()

    @property
    def auto_suspend_seconds(self):
        return self._auto_suspend_seconds

    def set_auto_suspend_seconds(self, value):
        self._auto_suspend_seconds = value
        self._update()

    @property
    def auto_resume(self):
        return self._auto_resume

    def set_auto_resume(self, value):
        self._auto_resume = value
        self._update()

    @property
    def instance_type_id(self):
        return self._instance_type_id

    @property
    def template_id(self):
        return self._template_id

    @property
    def environment_id(self):
        return self._environment_id

    def set_template_id(self, value):
        self._template_id = value
        self._update()

    def _get(self):
        return self.session.get_json(f"/engines/{self.id}")

    @property
    def status(self):
        return self._get()["engineStatus"]

    def __dict__(self):
        return {
            "id": self.id,
            "type": self.type,
            "name": self.name,
            "autoSuspendSeconds": self.auto_suspend_seconds,
            "autoResume": self.auto_resume,
            "templateId": self.template_id,
            "environmentId": self.environment_id,
            "serviceType": self.service_type,
        }

    def _update(self):
        resp = self.session.put_json(f"/engines/{self.id}", self.__dict__())
        return resp

    def start(self):
        if not self.session.is_plan_expired():
            try:
                self.session.post_json("/engines/start", self.__dict__())
                resp = self.session.get_json_until(
                    f"/engines/{self.id}",
                    lambda resp: resp["engineStatus"] == "active",
                    lambda resp: "Engine failed to start with status ERRORED"
                    if resp["engineStatus"] == "errored"
                    else None,
                )
                return resp
            except RuntimeError:
                logging.info("Engine is already active")
        else:
            raise PermissionError(
                "Engines are locked for expired plans. Contact us to upgrade.",
            )

    def stop(self):
        try:
            self.session.post_json("/engines/stop", self.__dict__())
            resp = self.session.get_json_until(
                f"/engines/{self.id}",
                lambda resp: resp["engineStatus"] == "suspended",
                lambda resp: "Engine failed to start with status ERRORED"
                if resp["engineStatus"] == "errored"
                else None,
            )
            return resp
        except RuntimeError:
            logging.info("Engine is already suspended")



import logging
import os 
from warnings import warn

import semantic_version
from decorator import decorator

from predibase.version import __version__


logger = logging.getLogger(__name__)


class PredibaseError(Exception):
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class PredibaseResponseError(PredibaseError):
    def __init__(self, message: str, code: int):
        super().__init__(message)
        self.code = code

is_ci = bool(os.getenv("PREDIBASE_CI", ""))

@decorator
def warn_outdated_sdk(fn, *args, **kwargs):
    resp = fn(*args, **kwargs)
    server_release_version = resp.headers.get("X-Predibase-Release-Version", None)
    if server_release_version is not None and not is_ci:
        if server_release_version == "" or server_release_version == "staging":
            if not __version__.startswith("0.1.1+dev") and not __version__.startswith("v2999"):
                warn(
                    "Using a post-release / prod version of the SDK in staging can lead to unexpected behavior. "
                    "Consider installing from latest master."
                )
        else:
            server_release_semver = semantic_version.Version(server_release_version)
            sdk_semver = semantic_version.Version(__version__)
            if sdk_semver < server_release_semver and not __version__.startswith("0.1.1+dev"):
                warn(
                    f"Currently installed SDK is outdated. This can lead to bugs or unexpected behavior. "
                    f"Consider upgrading to the latest version. Installed: {__version__} Latest: "
                    f"{server_release_version}."
                )

    return resp



from __future__ import annotations

from predibase.pql import Session
from predibase.resource.model import create_model_repo, ModelRepo, get_model_repo


class Repos:
    def __init__(self, session: Session):
        self._session = session

    def create(self, name: str, description: str | None = None, exists_ok: bool = False) -> ModelRepo:
        return create_model_repo(
            self._session,
            name,
            description,
            engine=None,
            exists_ok=exists_ok,
        )

    def get(self, resource_id: str) -> ModelRepo:
        if resource_id.startswith("pb://repos/"):
            resource_id = resource_id[: len("pb://repos/")]

        return get_model_repo(self._session, resource_id)

    def delete(self, resource_id: str):
        repo = self.get(resource_id)
        endpoint = f"/models/repo/{repo.id}"
        return self._session.delete_json(endpoint)



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from predibase._errors import PredibaseError
from predibase.resource.llm.interface import _LLMDeployment, LLMDeployment, LLMDeploymentJob

if TYPE_CHECKING:
    from predibase import Predibase


class Deployment:
    def __init__(
        self,
        client: Predibase,
        name: str,
    ):
        self._client = client
        self._session = client._session

        self.name = name


class DedicatedDeployment(Deployment):
    pass


class DedicatedLLMDeployment(DedicatedDeployment):
    def _refresh(self) -> LLMDeployment:
        data = _LLMDeployment.from_dict(self._session.get_json(f"/llms/{self.name}"))
        return LLMDeployment(self._session, self.name, deployment_metadata=data)

    def is_ready(self) -> bool:
        deployment = self._refresh()
        return deployment.is_ready()

    def wait_ready(self, timeout: int = 1500) -> "DedicatedLLMDeployment":
        deployment = self._refresh()
        status = deployment.data.deployment_status.lower()

        if status == "active":
            return self

        if status in ("terminating", "canceled", "inactive", "failed"):
            raise PredibaseError(f"deployment {self.name} is in a {status} state and will not become ready")

        if status in ("queued", "initializing", "updating"):
            _ = LLMDeploymentJob(self.name, self._session).get()
            return self

    @functools.cached_property
    def model_name(self) -> str:
        deployment = self._refresh()
        return deployment.data.model_name


class ServerlessDeployment(Deployment):
    pass


class ServerlessLLMDeployment(ServerlessDeployment):
    def _refresh(self) -> LLMDeployment:
        data = _LLMDeployment.from_dict(self._session.get_json(f"/llms/{self.name}"))
        return LLMDeployment(self._session, self.name, deployment_metadata=data)

    def is_ready(self) -> bool:
        deployment = self._refresh()
        return deployment.is_ready()

    def wait_ready(self, timeout: int = 1500) -> "ServerlessLLMDeployment":
        deployment = self._refresh()
        status = deployment.data.deployment_status.lower()

        if status == "active":
            if not deployment.wait_for_ready(timeout_seconds=timeout):
                raise PredibaseError(f"deployment {self.name} did not become ready within the {timeout} second timeout")

        if status in ("terminating", "canceled", "inactive", "failed"):
            raise PredibaseError(f"deployment {self.name} is in a {status} state and will not become ready")

        if status in ("queued", "initializing", "updating"):
            _ = LLMDeploymentJob(self.name, self._session).get()
            return self

    @functools.cached_property
    def model_name(self) -> str:
        deployment = self._refresh()
        return deployment.data.model_name



import copy
import datetime
import os
import tempfile
import time
import warnings
import zipfile
from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from http import HTTPStatus
from threading import Event, Thread
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import deprecation
import pandas as pd
import requests
from dataclasses_json import config, dataclass_json, LetterCase
from dateutil import parser

from predibase.pql.api import ServerResponseError, Session
from predibase.pql.utils import get_results_df
from predibase.resource.config import ConfigSuggestion
from predibase.resource.dataset import DataField, Dataset

# if TYPE_CHECKING:  # Only imports the below statements during type checking
#     from predibase.dataset import Dataset
from predibase.resource.engine import Engine
from predibase.resource.llm import interface
from predibase.resource.user import User
from predibase.resource.viz import VisualizeType
from predibase.resource_util import build_model_repo
from predibase.util import get_url, log_info, spinner
from predibase.util.metrics import is_llm_model
from predibase.util.tensorboard import launch_tensorboard as launch_tb
from predibase.util.url import encode_url_param
from predibase.version import __version__

TRAIN_METRICS = "train_metrics"
VALIDATION_METRICS = "validation_metrics"
TEST_METRICS = "test_metrics"

MODEL_WEIGHTS = "model_weights"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRepo:
    session: Session
    _id: str = field(metadata=config(field_name="id"))
    _uuid: str = field(metadata=config(field_name="uuid"))
    _name: str = field(metadata=config(field_name="modelName"))
    _created: str = field(metadata=config(field_name="created"))
    _updated: str = field(metadata=config(field_name="updated"))
    _description: Optional[str] = field(metadata=config(field_name="description"), default=None)
    _root_id: Optional[str] = field(metadata=config(field_name="rootID"), default=None)
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)
    _parent_id: Optional[str] = field(metadata=config(field_name="parentID"), default=None)
    # Latest dataset
    _latest_config: Optional[dict] = field(metadata=config(field_name="latestConfig"), default=None)
    _latest_dataset_id: Optional[int] = field(metadata=config(field_name="datasetID"), default=None)
    _latest_dataset: Optional[Dataset] = field(metadata=config(field_name="dataset"), default=None)

    def __repr__(self):
        return (
            f"ModelRepo(id={self.id}, name={self._name}, description={self.description}, "
            f"latest_config={{...}}, "
            f"latest_dataset={self.latest_dataset}, "
            f"created={self._created}, updated={self._updated})"
        )

    def _build_model_with_backref(self, resp):
        from predibase.resource_util import build_model

        resp["repo"] = self
        return build_model(resp, self.session)

    @property
    def id(self):
        return self._id

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def latest_config(self):
        return self._latest_config

    @property
    def latest_dataset(self):
        if self._latest_dataset:
            return self._latest_dataset
        # get_model().repo.latest_dataset isn't seeded
        if self._latest_dataset_id:
            resp = self.session.get_json(f"/datasets/{self._latest_dataset_id}")
            from predibase.resource_util import build_dataset

            self._latest_dataset = build_dataset(resp, self.session)
            return self._latest_dataset
        return None

    @property
    def latest_dataset_id(self):
        return self._latest_dataset_id

    @property
    def created(self):
        return self._created

    @property
    def updated(self):
        return self._updated

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def root_id(self):
        return self._root_id

    @property
    def author(self):
        return self._user.username if self._user else None

    def create_draft(
        self,
        dataset: Optional[Dataset] = None,
        target: Optional[Union[str, List[str]]] = None,
        automl: Optional[bool] = False,
        hyperopt: Optional[bool] = False,
        config: Optional[dict] = None,
        description: Optional[str] = None,
    ) -> "ModelDraft":
        dataset = dataset or self.latest_dataset
        if dataset is None:
            raise ValueError(
                "Model repo has no latest dataset, and no dataset was provided. "
                "Cannot create model draft without a dataset.",
            )

        # need target to get the default config.
        if not config and target:
            config = get_default_config(self.session, dataset, target, automl, hyperopt)
        return ModelDraft(
            session=self.session,
            repo_id=self.id,
            _repo=self,
            name=self._name,
            config=config,
            dataset=dataset,
            description=description,
        )

    def _filter_model_weights(self, filename: str) -> bool:
        return "readme" not in filename.lower()

    def upload_model(self, model_path: str, dataset: Optional[Dataset] = None) -> "Model":
        required_metadata_files = ["model_hyperparameters.json", "training_progress.json", "training_set_metadata.json"]

        dataset_id = dataset.id if dataset is not None else self.latest_dataset_id
        if dataset_id is None:
            raise ValueError("must either provide a dataset or upload to a model repo with a dataset configured")

        with tempfile.TemporaryDirectory() as tmpdir:
            # If the path specified is a local folder, no need to unzip.
            model_artifacts_path = model_path
            if not os.path.isdir(model_path):
                model_artifacts_path = tmpdir
                with zipfile.ZipFile(model_path, "r") as zf:
                    zf.extractall(path=tmpdir)

            # Some model zips may save the files of interest in a model/ sub-directory.
            if os.path.exists(os.path.join(model_artifacts_path, "model")):
                model_artifacts_path = os.path.join(model_artifacts_path, "model")

            for required_file in required_metadata_files + [MODEL_WEIGHTS]:
                if not os.path.exists(os.path.join(model_artifacts_path, required_file)):
                    raise FileNotFoundError(
                        f"required file {required_file} not found in model folder {model_artifacts_path}",
                    )

            if not os.path.exists(os.path.join(model_artifacts_path, MODEL_WEIGHTS)):
                raise FileNotFoundError(f"model_weights file(s) not found in model folder {model_artifacts_path}")

            model_weights_files = []
            if os.path.isdir(os.path.join(model_artifacts_path, MODEL_WEIGHTS)):
                for file in os.listdir(os.path.join(model_artifacts_path, MODEL_WEIGHTS)):
                    if self._filter_model_weights(file):
                        model_weights_files.append(file)
            else:
                model_weights_files = [MODEL_WEIGHTS]

            metadata_zip_path = os.path.join(tmpdir, "model_metadata.zip")
            with zipfile.ZipFile(metadata_zip_path, "w") as metadata_zip:
                for f in required_metadata_files:
                    metadata_zip.write(os.path.join(model_artifacts_path, f), arcname=f)

            with open(metadata_zip_path, "rb") as zipped_metadata:
                files = {"file": zipped_metadata}
                file_name, _ = os.path.splitext(os.path.basename(metadata_zip_path))
                s = ","
                data = {
                    "fileName": file_name,
                    "datasetID": dataset_id,
                    "repoID": self.id,
                    "modelWeightsFiles": s.join(model_weights_files),
                }
                register_model_resp = self.session.post("/models/register_uploaded_model", files=files, data=data)

            model_dict = register_model_resp["model"]
            presigned_urls = register_model_resp["modelWeightsPresignedFiles"]

            # Loop through the model weight files and upload them to blob storage.
            # Beacuse we send the list of files to upload to the backend, we should
            # be able to rely on the order of the response matching our list.
            for idx, model_weight_file in enumerate(model_weights_files):
                presigned_url = presigned_urls[idx]
                upload_url = presigned_url["url"]
                upload_required_headers = presigned_url.get("headers", {})
                if upload_required_headers is None:
                    upload_required_headers = {}

                content_type = (
                    "application/json" if model_weight_file.lower().endswith(".json") else "binary/octet-stream"
                )
                headers = {"Content-Type": content_type}
                for k, v in upload_required_headers:
                    headers[k] = v

                file_path = (
                    os.path.join(model_artifacts_path, model_weight_file)
                    if model_weight_file == MODEL_WEIGHTS
                    else os.path.join(model_artifacts_path, "model_weights", model_weight_file)
                )
                with open(file_path, "rb") as weights_file:
                    # Upload file to blob storage with pre-signed url
                    requests.put(upload_url, data=weights_file, headers=headers).raise_for_status()

            # Signal successful upload of model weights.
            self.session.post(f"/models/complete_upload_model/{model_dict['id']}")

            model_dict["status"] = "ready"
            return self._build_model_with_backref(model_dict)

    def list_models(self, df: bool = False):
        endpoint = f"/models/repo/{self._id}?withVersions=true"
        resp = self.session.get_json(endpoint)
        if df:
            return pd.DataFrame([m for m in resp["modelRepo"]["models"]])
        return [self._build_model_with_backref(m) for m in resp["modelRepo"]["models"]]

    def draft_from_latest_version(self) -> "ModelDraft":
        return ModelDraft(
            session=self.session,
            repo_id=self.id,
            _repo=self,
            name=self._name,
            config=copy.deepcopy(self.latest_config),
            dataset=self.latest_dataset,
        )

    def latest(self) -> "Model":
        return get_model(self.session, self.name)

    def head(self) -> "Model":
        return self.latest()

    def get(self, version: Optional[Union[int, str]] = None) -> "Model":
        return get_model(self.session, self.name, version)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunMetadata:
    _id: str = field(metadata=config(field_name="id"))
    _model_id: int = field(metadata=config(field_name="modelID"))
    _run_id: str = field(metadata=config(field_name="runID"))
    _rendered_config: dict = field(metadata=config(field_name="renderedConfig"))

    @property
    def rendered_config(self) -> dict:
        return self._rendered_config


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Model:
    # Disable the dataclass-json violation (ref: https://predibase.slack.com/archives/C02AZ1USR1U/p1696446768000669)
    # pylint: disable=used-before-assignment
    session: Session
    _id: str = field(metadata=config(field_name="id"))
    _uuid: str = field(metadata=config(field_name="uuid"))
    _repo_id: int = field(metadata=config(field_name="repoID"))
    _repo: ModelRepo = field(metadata=config(field_name="repo"))
    _config: dict = field(metadata=config(field_name="config"))
    _dataset_id: int = field(metadata=config(field_name="datasetID"))
    _dataset: Dataset = field(metadata=config(field_name="dataset"))
    _version: str = field(metadata=config(field_name="repoVersion"))
    _status: str = field(metadata=config(field_name="status"))
    _starred: bool = field(metadata=config(field_name="starred"))
    _archived: bool = field(metadata=config(field_name="archived"))
    # _created: datetime.datetime = config(metadata=config(field_name="created"))
    _created: str = field(metadata=config(field_name="created"))
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)
    _created_by_user_id: Optional[int] = field(metadata=config(field_name="createdByUserID"), default=None)
    _engine_id: Optional[str] = field(metadata=config(field_name="engineID"), default=None)
    _engine: Optional[Engine] = field(metadata=config(field_name="engine"), default=None)
    _engine_template: Optional[Dict[str, Any]] = field(metadata=config(field_name="engineTemplate"), default=None)
    _description: Optional[str] = field(metadata=config(field_name="description"), default=None)
    _experiment_id: Optional[int] = field(metadata=config(field_name="experimentID"), default=None)
    _error_text: Optional[str] = field(metadata=config(field_name="errorText"), default=None)
    _best_run_id: Optional[str] = field(metadata=config(field_name="bestRunID"), default=None)
    _active_run_id: Optional[str] = field(metadata=config(field_name="activeRunID"), default=None)
    _completed: Optional[str] = field(metadata=config(field_name="completed"), default=None)
    _training_created: Optional[str] = field(metadata=config(field_name="trainingCreated"), default=None)
    _training_completed: Optional[str] = field(metadata=config(field_name="trainingCompleted"), default=None)
    _active_run_meta: Optional[ModelRunMetadata] = field(metadata=config(field_name="modelRun"), default=None)
    _llm_base_model_name: Optional[str] = field(metadata=config(field_name="llmBaseModelName"), default=None)

    def __repr__(self):
        dataset = f"Dataset({self._dataset.name}...)"
        repo = f"Repo({self.repo.name}...)"
        return (
            f"Model(id={self._id}, repo={repo}, description={self.description}, "
            f"dataset={dataset}, engine=Engine({self.engine.name if self.engine else None}...), "
            f"config={{...}}, version={self.version}, "
            f"status={self.status}, created={self.created}, completed={self.completed})"
        )

    @property
    def id(self):
        return self._id

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def best_run_id(self) -> str:
        return self._best_run_id

    @property
    def description(self):
        return self._description

    @property
    def config(self):
        return self._config

    @property
    def rendered_config(self):
        if self._active_run_meta is None:
            return None
        return self._active_run_meta.rendered_config

    @property
    def version(self):
        return self._version

    @property
    def repo_id(self):
        return self._repo_id

    @property
    def author(self):
        return self._user.username if self._user else None

    @property
    def repo(self):
        return self._repo

    @property
    def dataset_id(self):
        return self._dataset_id

    @property
    def engine_id(self):
        return self._engine_id

    @property
    def engine(self):
        if self._engine:
            return self._engine
        resp = self.session.get_json(f"/engines/{self.engine_id}")
        engine = Engine.from_dict({"session": self.session, **resp})
        self._engine = engine
        return engine

    @property
    def engine_template(self):
        return self._engine_template

    @property
    def dataset(self):
        return self._dataset

    @property
    def status(self):
        return self._get()["status"]

    @property
    def error_text(self) -> Union[str, None]:
        resp = self._get()
        if "errorText" in resp:
            return resp["errorText"]
        return None

    @property
    def active_fields(self) -> List[DataField]:
        resp = self._get()
        return [DataField.from_dict({"session": self.session, **x["field"]}) for x in resp["activeFields"]]

    @property
    def starred(self) -> bool:
        return self._starred

    @property
    def archived(self) -> bool:
        return self._archived

    @property
    def created(self) -> datetime.datetime:
        return pd.to_datetime(self._created)

    @property
    def completed(self) -> datetime.datetime:
        if self._completed:
            return pd.to_datetime(self._completed)
        resp = self._get()
        if "completed" in resp:
            self._completed = resp["completed"]
            return pd.to_datetime(self._completed)
        return None

    @property
    def training_created(self) -> datetime.datetime:
        if self._training_created:
            return self._training_created
        resp = self._get()
        if "trainingCreated" in resp:
            self._training_created = resp["trainingCreated"]
            return self._training_created
        return None

    @property
    def training_completed(self) -> datetime.datetime:
        if self._training_completed:
            return self._training_completed
        resp = self._get()
        if "trainingCompleted" in resp:
            self._training_completed = resp["trainingCompleted"]
            return self._training_completed
        return None

    @cached_property
    def metrics(self) -> "ModelMetrics":
        nested_metrics = {
            TRAIN_METRICS: defaultdict(dict),
            VALIDATION_METRICS: defaultdict(dict),
            TEST_METRICS: defaultdict(dict),
        }

        global_metrics = {
            TRAIN_METRICS: defaultdict(dict),
            VALIDATION_METRICS: defaultdict(dict),
            TEST_METRICS: defaultdict(dict),
        }

        general_metrics = {}

        for metric in self.active_run.data.metrics:
            parts = metric.key.split(".")

            if len(parts) == 1:
                general_metrics[metric.key] = metric.value
                continue

            # we'd like to capture the best metrics of the run.
            if parts[0] != "best" or len(parts) < 2:
                continue

            # if this field contains metrics, its format will look like the following
            # "best.train_metrics.Survived.accuracy"
            _, split, feature, name = parts

            if feature != "combined":
                nested_metrics[split][feature][name] = metric.value
            else:
                global_metrics[split][feature][name] = metric.value

        return ModelMetrics(nested_metrics, global_metrics, general_metrics)

    @property
    def active_run(self):
        return self.runs[self._active_run_id]

    @cached_property
    def runs(self) -> Dict[str, "ModelRun"]:
        resp = self.session.get_json(f"/models/version/{self.id}?withRuns=true")
        runs: List[ModelRun] = [ModelRun.from_dict(run) for run in resp["runs"]]
        return {run.info.run_id: run for run in runs}

    @property
    def llm_base_model_name(self) -> Optional[str]:
        return self._llm_base_model_name

    def _get(self):
        endpoint = f"/models/version/{self.id}"
        resp = self.session.get_json(endpoint)
        return resp["modelVersion"]

    def update_description(self, description: str):
        endpoint = f"/models/version/{self._id}"
        self.session.put_json(
            endpoint,
            {
                "id": self.id,
                "description": description,
                "starred": self.starred,
                "archived": self.archived,
            },
        )

    def delete(self):
        return self.session.delete_json(f"/models/version/{self.id}")

    def cancel(self):
        return self.session.post_json(f"/models/version/{self.id}/cancel", {})

    def _model_download_endpoint(self, framework: str) -> str:
        endpoint = f"/models/download/{self.id}"
        return endpoint + f"?framework={framework.lower()}"

    @spinner(name="exporting model")
    def _poll_model_download(self, endpoint: str) -> Dict[str, Any]:
        while True:
            resp = self.session.get_json(endpoint)
            status = resp["status"]
            if status == "exporting":
                time.sleep(2)
            elif status == "complete":
                break
            else:
                raise ValueError(f"Error exporting model: {resp['failure_reason']}")

        return resp

    def _start_model_download(self, endpoint: str) -> Dict[str, Any]:
        try:
            self.session.post(endpoint, {})
        except ServerResponseError as e:
            raise e

        return self._poll_model_download(endpoint)

    @deprecation.deprecated(
        deprecated_in="2024.2.12",
        current_version=__version__,
        details="framework parameter is deprecated",
    )
    def download(
        self,
        name: Optional[str] = None,
        location: Optional[str] = None,
        framework: Optional[str] = None,
    ):
        """
        This method downloads the model to the specified location on your local machine.
        Args:
            name: Optional[str] - Name for the downloaded file.
            location: Optional[str] - File path to download the model to.
            framework: Optional[str] - Framework to download the model as. Default is to download as a Ludwig Model.

        Returns:
            None
        """
        if framework is None:
            framework = "ludwig"
        else:
            warnings.warn("framework parameter is deprecated and will be removed in an upcoming release.")

        if is_llm_model(self._config) and framework != "ludwig":
            raise ValueError("LLMs can only be exported using the Ludwig format.")

        if location is not None:
            if not os.path.exists(location):
                raise ValueError(f"Location {location} does not exist.")

        endpoint = self._model_download_endpoint(framework)

        log_info("Downloading model")
        try:
            resp = self.session.get_json(endpoint)
        except ServerResponseError as e:
            # Model has not been previously exported. Start process.
            if "error 400: record not found" in e.message.lower():
                log_info("Model has not been exported yet. Starting export.")
                resp = self._start_model_download(endpoint)
                pass
            else:
                raise e

        status = resp["status"]
        # Wait for model to finish exporting
        if status == "exporting":
            resp = self._poll_model_download(endpoint)
        # If model was previously exported but failed, restart process
        elif status == "failed":
            resp = self._start_model_download(endpoint)

        # Now that model is successfully exported, download it
        url = resp["url"]
        download_path = None
        if name:
            if name[-4:] == ".zip":
                download_path = name
            else:
                download_path = name + ".zip"
        else:
            download_path = os.path.basename(urlparse(url).path)

        if location is not None:
            download_path = os.path.join(location, download_path)

        resp = requests.get(url)
        with open(download_path, "wb") as f:
            f.write(resp.content)

        log_info(f"Model downloaded to {download_path}")

    @deprecation.deprecated(
        deprecated_in="2024.2.12",
        current_version=__version__,
        details="framework parameter is deprecated",
    )
    def export(
        self,
        location: str,
        connection: Optional[Union[str, int]] = None,
        framework: Optional[str] = None,
    ):
        """Export a model to an export location using the connection or target connection with the optional
        framework.

        :param location: Location as string.
        :param connection: (Optional[str, int]): Optional connection id or name for export location.
        :param framework: Optional framework, defaults to "ludwig"
        """
        connection_id = self.session.get_connection_id(connection) if connection else None
        query = f"EXPORT MODEL {self.repo.name} VERSION {self.version}"
        if framework is not None:
            warnings.warn("framework parameter is deprecated and will be removed in an upcoming release.")
            query = query + f" WITH FORMAT (framework='{framework}')"
        query = query + f" TO '{location}'"
        return self.session.execute(query, connection_id=connection_id)

    def evaluate(
        self,
        targets: Union[str, List[str]],
        source: Union[Dataset, pd.DataFrame, str],
        limit: Optional[int] = None,
        metadata: Optional[bool] = False,
    ) -> pd.DataFrame:
        """Evaluate a model given a source dataset for one or more targets.

        :param targets (Union[str, List[str], Dict[str, str]]): One or more target names to make predictions on.
        :param source (Union[Dataset, pd.DataFrame, str]): Source dataset to evaluate.
        :param metadata (Optional[bool]): Optional boolean flag to return model name and version metadata in results.
        """
        if isinstance(targets, list):
            predict_targets = ", ".join([f'"{t}"' for t in targets])
        else:
            predict_targets = f'"{targets}"'

        # create list of properties
        properties = []
        if metadata:
            properties.append("METADATA")

        if len(properties) > 0:
            predict_targets += " WITH " + ", ".join(properties)

        # Add using query for current model and the latest version
        model_name = f'"{self.repo.name}" VERSION {self.version}'

        # Source can be a dataset, dataframe, or query string.
        source, params, connection_id = self._build_source(source, limit)

        # Set session connection from dataset if not set
        if not self.session.connection_id and self.dataset:
            self.session.set_connection(self.dataset.connection_id)
        query = f"EVALUATE {predict_targets} USING {model_name} GIVEN {source}"
        return self.session.execute(query, params=params, connection_id=connection_id)

    def predict(
        self,
        targets: Union[str, List[str], Dict[str, str]],
        source: Union[Dataset, pd.DataFrame, str],
        limit: Optional[int] = None,
        explanation: Optional[Union[bool, str]] = False,
        confidence: Optional[bool] = False,
        probabilities: Optional[Union[bool, List[any]]] = False,
        metadata: Optional[bool] = False,
        engine: Optional[Engine] = None,
    ) -> pd.DataFrame:
        """Make predictions for one or more target variables given a source dataset.

        :param targets: (Union[str, List[str], Dict[str, str]]): One or more target names to make predictions on.
        Include a dictionary with key being the target and the value being an alias to return in results.
        :param source: (Union[Dataset, pd.DataFrame, str]): Source to perform predictions on.
        :param limit: Number of rows to return.
        :param explanation: (Optional[Union[bool, str]]): Optional flag to enable explanations.
        Provide a string to return explanations for a specific algorithm.  Supported values are 'ig' or 'shap'.
        :param confidence: (Optional[bool]): Optional boolean flag to return metadata in results.
        :param probabilities: (Optional[Union[bool, List[any]]]): Optional flag to return probabilities for binary of
            multi-class classificaiton.
        Provide a list of category names eg [True,False] to include a specific set of classes.
        :param metadata: (Optional[bool]): Optional boolean flag to return model name and version metadata in results.
        :param engine: (Optional[Engine]): Option to select a particular engine for executing the predict.
        """

        # Add predict targets with optional alias, list, or as simple string
        if isinstance(targets, dict):
            predict_targets = ", ".join([f'"{k}" AS "{v}"' for (k, v) in targets.items()])
        elif isinstance(targets, list):
            predict_targets = ", ".join([f'"{t}"' for t in targets])
        else:
            predict_targets = f'"{targets}"'

        # create list of properties
        properties = []
        if confidence:
            properties.append("CONFIDENCE")
        if explanation:
            # TODO: Also add an enum type here for explanation instead of just string
            if isinstance(explanation, str):
                properties.append(f"EXPLANATION (algo='{explanation}')")
            else:
                properties.append("EXPLANATION")
        if probabilities:
            categories = ""
            if isinstance(probabilities, list):
                categories = ", ".join([f"'{str(c)}'" for c in probabilities])
                categories = f" OF ({categories})"
            properties.append("PROBABILITIES" + categories)
        if metadata:
            properties.append("METADATA")

        if len(properties) > 0:
            predict_targets += " WITH " + ", ".join(properties)

        # Add using query for current model and the latest version
        model_name = f'"{self.repo.name}" VERSION {self.version}'

        # Source can be a dataset, dataframe, or query string.
        source, params, connection_id = self._build_source(source, limit)
        # If using a dataframe, use model.dataset.connection_id
        if params:
            connection_id = self.dataset.connection_id

        query = f"PREDICT {predict_targets} USING {model_name} GIVEN {source}"
        return self.session.execute(
            query,
            params=params,
            connection_id=connection_id,
            engine_id=engine.id if engine else None,
        )

    def deploy(
        self,
        deployment_name: str,
        engine_template: Optional[str] = None,
        hf_token: Optional[str] = None,
        auto_suspend_seconds: Optional[int] = None,
        max_input_length: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        max_batch_prefill_tokens: Optional[int] = None,
        revision: Optional[str] = None,
        quantization_kwargs: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False,
    ) -> "interface.LLMDeploymentJob":
        if not self._llm_base_model_name:
            raise NotImplementedError(
                "Support for non-LLM deployments via this API is not yet implemented. "
                "Please use predibase_client.create_deployment() instead",
            )

        return interface.deploy_llm(
            session=self.session,
            deployment_name=deployment_name,
            engine_template=engine_template,
            hf_token=hf_token,
            auto_suspend_seconds=auto_suspend_seconds,
            max_input_length=max_input_length,
            max_total_tokens=max_total_tokens,
            max_batch_prefill_tokens=max_batch_prefill_tokens,
            quantization_kwargs=quantization_kwargs,
            revision=revision,
            predibase_model_uri=f"pb://models/{self.repo.name}/{self.version}",
            overwrite=overwrite,
        )

    def visualize(
        self,
        visualize_type: Union[VisualizeType, str],
        output_feature: Optional[str] = None,
        plot_images: Optional[bool] = True,
    ) -> List[str]:
        self._wait_until_ready()

        if isinstance(visualize_type, VisualizeType):
            visualize_type = visualize_type.value
        endpoint = f"/visualize?modelID={self.id}&visualizeType={visualize_type}"
        if output_feature:
            endpoint += f"&outputFeature={output_feature}"
        resp = self.session.get_json(endpoint)
        if plot_images:
            import ipyplot

            ipyplot.plot_images(
                resp["links"],
                img_width=400,
            )
        return resp["links"]

    def get_evaluation_statistics(self):
        self._wait_until_ready()
        endpoint = f"/visualize?modelID={self.id}&visualizeType=evaluation_statistics"
        resp = self.session.get_json(endpoint)
        return resp["testStatistics"]

    def get_feature_importance(self, output_feature: Optional[str] = None) -> pd.DataFrame:
        self._wait_until_ready()

        if output_feature is None:
            output_features = [x["name"] for x in self.config["output_features"]]
            if len(output_features) == 1:
                output_feature = output_features[0]
            else:
                raise ValueError(
                    "No output feature provided and "
                    "cannot be inferred because model contains multiple output features",
                )
        endpoint = f"/visualize/explanations?modelID={self.id}&outputFeature={output_feature}"
        resp = self.session.get_json(endpoint)
        return get_results_df({"dataset": resp})

    def get_training_timeline(self) -> pd.DataFrame:
        """
        :return: Dataframe with columns [Step, Start, End, Duration], where Step in
            {Preprocessing, Training, Evaluating, etc.}
        """
        endpoint = f"/models/version/{self.id}/timeline"
        resp = self.session.get_json_until(
            endpoint=endpoint,
            success_cond=lambda resp: all(
                [resp[process]["created"] is not None and resp[process]["completed"] is not None for process in resp],
            ),
        )
        columns = ["step", "error_message", "start", "end", "duration"]
        data = []
        for process in resp:
            start = parser.parse(resp[process]["created"])
            end = parser.parse(resp[process]["completed"])
            error_message = resp[process].get("errorMessage", None)
            data.append([process, error_message, start, end, end - start])
        return pd.DataFrame(columns=columns, data=data)

    def _build_source(
        self,
        source: Union[Dataset, pd.DataFrame, str],
        limit: Optional[int] = None,
    ) -> Tuple[str, Optional[str], Optional[str]]:
        # Source can be a dataset, dataframe, or query string.
        params = None
        connection_id = None
        if isinstance(source, Dataset):
            connection_id = source.connection_id
            source = f'SELECT * FROM "{source.connection.name}"."{source.name}"'
            if limit is not None:
                source += f" LIMIT {limit}"
        elif isinstance(source, pd.DataFrame):
            params = {"sdk_df_dataset": source.to_json()}
            source = "sdk_df_dataset"
        elif not isinstance(source, str):
            raise Exception(f"unexpected source type {type(source)}")
        return source, params, connection_id

    def to_draft(self):
        return ModelDraft(
            session=self.session,
            name=self.repo.name,
            repo_id=self.repo_id,
            _repo=self._repo,
            config=copy.deepcopy(self.config),
            dataset=self.dataset,
        )

    def fork(self, name):
        return ModelDraft(
            session=self.session,
            name=name,
            config=copy.deepcopy(self.config),
            dataset=self.dataset,
            parent_id=self.repo_id,
        )

    def get_active_fields(self) -> List[DataField]:
        """
        Queries like "PREDICT <field>> GIVEN SELECT * FROM <dataset>" that don't specify an explicit model
        will use this model for each of the fields returned here
        """
        return self.active_fields

    def activate_model_for_field(self, field: DataField):
        """Sets this model as the default model used for operations on a data field. For example, if you have two
        models.

        [model_tabnet, model_concat] trained on the same dataset, you can call
        model_tabnet.set_default_for_field(DataField(id=..., name="Survived")).

        Queries like "PREDICT <field>> GIVEN SELECT * FROM <dataset>" that don't specify an explicit model
        will use the default model set here for the "Survived" field

        :param field: DataField of dataset. Obtained from dataset.get_fields()
        """
        if field.id in [x.id for x in self.active_fields]:
            raise ValueError(f"field {field.name} is already active for this model")
        return self.session.post_json("/models/activate", {"modelID": self.id, "fieldIDs": [field.id]})

    def deactivate_model_for_field(self, field: DataField):
        if field.id not in [x.id for x in self.active_fields]:
            raise ValueError(f"field {field.name} is not active for this model")
        return self.session.post_json("/models/activate", {"modelID": self.id, "fieldIDs": [field.id]})

    def _wait_until_ready(self, launch_tensorboard=False):
        """Utility function to wait until a model is fully ready, including evaluation and visualization steps.
        Certain operations may otherwise fail in a racy way.

        :param model: the Model to wait for.
        :return: None
        """
        self.session._wait_until_model_ready_with_logging(self._id, launch_tensorboard)

    def launch_tensorboard(self):
        # Check if the model has logs available or will have logs available
        model_logs = self.session.get_model_logs_urls(self._id)
        if len(model_logs.get("logfiles", [])) == 0 and not model_logs.get("moreLogsAvailable", False):
            print(
                f"Model {self._id} does not have logs available, which is likely because the model failed or was terminated before logs were created. Please check model status.",  # noqa: E501
            )
            return

        # Sync files in subthread and launch TB in main thread
        tensorboard_event = Event()
        logs_sync_thread = Thread(
            target=self.session.get_model_logs,
            args=[self._id, tensorboard_event, None],
            daemon=True,
        )
        logs_sync_thread.start()
        launch_tb(self._id)

        # Stop syncing files when TB is stopped
        tensorboard_event.set()
        logs_sync_thread.join()


class ModelFuture:
    def __init__(self, version_id: int, session: Session):
        self.version_id = version_id
        self.session = session

    def wait(self):
        self._wait_for_ready()

    def get(
        self,
        launch_tensorboard: bool = False,
    ) -> Model:
        log_info("Monitoring status of model training...")

        for _ in range(5):
            resp = self.session.get_json(f"/models/version/{self.version_id}")
            engine_template = resp.get("modelVersion", {}).get("engineTemplate", None)

            if engine_template is None:
                time.sleep(1)
                continue

            cloud = next(iter(engine_template["headNodeGroup"]["nodeType"]["clouds"]))

            node_counts = defaultdict(int)
            head_node_group_cloud = engine_template["headNodeGroup"]["nodeType"]["clouds"][cloud]
            head_node_label = head_node_group_cloud.get("publicLabel", head_node_group_cloud["label"])
            node_counts[head_node_label] += 1
            for worker_node_group in engine_template.get("workerNodeGroups", []):
                public_label = (
                    worker_node_group.get("nodeType", {}).get("clouds", {}).get(cloud, {}).get("publicLabel", "")
                )
                label = (
                    public_label
                    if public_label
                    else worker_node_group.get("nodeType", {}).get("clouds", {}).get(cloud, {}).get("label", "")
                )
                count = int(
                    worker_node_group.get("replicas", "0"),
                )
                if label:
                    node_counts[label] += count

            log_info("Compute summary:")
            for label, count in sorted(node_counts.items()):
                if count > 0:
                    log_info(f"  * {label} (x{count})")

            break

        if launch_tensorboard:
            launch_tb(self.version_id)

        resp = self._wait_for_ready(
            launch_tensorboard=launch_tensorboard,
        )

        from predibase.resource_util import build_model

        return build_model(resp["modelVersion"], self.session)

    def cancel(self):
        self.session.post_json(f"/models/version/{self.version_id}/cancel", {})

    def is_finished(self) -> bool:
        resp = self.session.get_json(f"/models/version/{self.version_id}")
        return resp["modelVersion"]["status"] in {"ready", "failed", "canceled"}

    def _wait_for_ready(self, launch_tensorboard=False) -> Dict[str, Any]:
        return self.session._wait_until_model_ready_with_logging(self.version_id, launch_tensorboard)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelDraft:
    session: Session
    dataset: Dataset
    config: dict
    description: Optional[str] = None
    repo_id: Optional[int] = None
    _repo: Optional[ModelRepo] = None
    name: Optional[str] = None
    parent_id: Optional[str] = None

    def __repr__(self):
        repo_str = f"repo=({self.repo.name}...) " if self.repo else "repo=(<new repo from fork>)"
        return (
            f"ModelDraft(name={self.name}, dataset=({self.dataset.name}...), config={{...}}, "
            f"description={self.description}, {repo_str}, parent_id={self.parent_id})"
        )

    @property
    def repo(self):
        return self._repo

    # Removed https://github.com/predibase/predibase/pull/3770
    # @spinner(name="Train Model")
    def train(self, engine: Optional[Engine] = None, launch_tensorboard: bool = False) -> Model:
        if not self.session.is_plan_expired():
            future = self.train_async(engine)
            return future.get(launch_tensorboard=launch_tensorboard)
        else:
            raise PermissionError(
                "Training models is locked for expired plans. Contact us to upgrade.",
            )

    def train_async(self, engine: Optional[Engine] = None) -> ModelFuture:
        if not self.session.is_plan_expired():
            # Retraining
            if self.repo_id:
                request = {
                    "config": self.config,
                    "datasetID": self.dataset.id,
                    "repoID": self.repo_id,
                    "engineID": engine.id if engine is not None else None,
                    "description": self.description,
                }

                resp = self.session.post_json("/models/train", request)
                model_id = resp["model"]["id"]

                # Output link to UI model page
                endpoint = f"models/version/{model_id}"
                url = get_url(self.session, endpoint)
                log_info(f"Check Status of Model Training Here: [link={url}]{url}")

                return ModelFuture(model_id, self.session)

            # Forking
            endpoint = "/models/repo"
            request = {
                "modelType": "ludwig",
                "modelName": self.name,
                "latestConfig": self.config,
                "retrainCadence": "",
                "retrainConfig": {},
                "datasetID": self.dataset.id,
                "parentID": self.parent_id,
            }
            resp = self.session.post_json(endpoint, request)
            repo_id = resp["id"]
            log_info(f"Created model repository: <{self.name}>")

            resp = self.session.post_json(
                "/models/train",
                {
                    "datasetID": self.dataset.id,
                    "config": self.config,
                    "repoID": repo_id,
                    "engineID": engine.id if engine is not None else None,
                },
            )
            model_id, model_version = resp["model"]["id"], resp["model"]["repoVersion"]
            log_info(f"Training model version {model_version} for model repository <{self.name}>...")
            return ModelFuture(model_id, self.session)
        else:
            raise PermissionError(
                "Training models is locked for expired plans. Contact us to upgrade.",
            )


class FeatureMetrics:
    def __init__(self, values: Dict[str, float]):
        for k, v in values.items():
            setattr(self, k, v)


class SplitMetrics:
    def __init__(
        self,
        nested_metrics: Dict[str, Dict[str, float]],
        global_metrics: Dict[str, float],
    ):
        for k, v in nested_metrics.items():
            setattr(self, k, FeatureMetrics(v))
        for k, v in global_metrics.items():
            setattr(self, k, v)


class ModelMetrics:
    def __init__(
        self,
        nested_metrics: Dict[str, Dict[str, Dict[str, float]]],
        global_metrics: Dict[str, Dict[str, float]],
        general_metrics: Dict[str, int],
    ):
        # General metrics includes things like epochs, batch_size, etc.
        self.general_metrics = general_metrics
        self.train = SplitMetrics(nested_metrics[TRAIN_METRICS], global_metrics[TRAIN_METRICS])
        self.validation = SplitMetrics(nested_metrics[VALIDATION_METRICS], global_metrics[VALIDATION_METRICS])
        self.test = SplitMetrics(nested_metrics[TEST_METRICS], global_metrics[TEST_METRICS])


# TODO(travis): Consider moving these off of MLFlow format
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunMetric:
    key: str
    step: int
    timestamp: float
    value: float


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunParam:
    key: str
    value: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunTag:
    key: str
    value: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunInfo:
    artifact_uri: str
    experiment_id: str
    lifecycle_stage: str  # active
    run_id: str
    run_uuid: str
    status: str
    user_id: str
    end_time: Optional[float] = field(metadata=config(field_name="EndTime"), default=None)
    start_time: Optional[float] = field(metadata=config(field_name="StartTime"), default=None)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunData:
    metrics: List[ModelRunMetric]
    params: List[ModelRunParam]
    tags: List[ModelRunTag]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRun:
    # https://www.mlflow.org/docs/latest/_modules/mlflow/entities/run_data.html
    data: ModelRunData

    # https://www.mlflow.org/docs/latest/_modules/mlflow/entities/run_info.html#RunInfo
    info: ModelRunInfo


def get_suggested_configs(
    session: Session,
    base_config: Dict[Any, str],
) -> List[ConfigSuggestion]:
    """Gets config suggestions given a base config."""
    resp = session.post_json(
        "/config/suggest",
        {
            "config": base_config,
        },
    )
    config_suggestions = resp["configs"]
    return [ConfigSuggestion.from_dict(suggestion) for suggestion in config_suggestions]


def get_default_config(
    session: Session,
    dataset: Dataset,
    targets: Optional[Union[str, List[str]]] = None,
    automl: Optional[bool] = False,
    hyperopt: Optional[bool] = False,
) -> dict:
    endpoint = f"/config/detect/{dataset.id}"
    resp = session.get_json(endpoint)
    config = resp["config"]

    if targets:
        if isinstance(targets, str):
            targets = [targets]
        output_features = [x for x in config["input_features"] if x["name"] in targets]
        config["input_features"] = [x for x in config["input_features"] if x["name"] not in targets]
        config["output_features"] = output_features
    if not automl:
        if "combiner" in config:
            del config["combiner"]
        if "trainer" in config:
            del config["trainer"]
    if not hyperopt:
        if "hyperopt" in config:
            del config["hyperopt"]
    return config


def get_model(
    session: Session,
    name: Optional[str] = None,
    version: Optional[Union[int, str]] = None,
    model_id: Optional[int] = None,
) -> Model:
    if not (name or model_id):
        raise ValueError("Must provide either model_name or model_id")
    if name and model_id:
        raise ValueError("Cannot provide both model name and model_id")
    if version:
        if model_id:
            raise ValueError("Cannot provide both model version and model_id")

    if model_id:
        endpoint = f"/models/version/{model_id}"
        resp = session.get_json(endpoint)["modelVersion"]
    else:
        endpoint = f"/models/version/name/{name}"
        endpoint += f"?version={version}" if version else ""
        resp = session.get_json(endpoint)

    from predibase.resource_util import build_model

    return build_model(resp, session)


def get_model_repo(session: Session, name: str) -> ModelRepo:
    resp = session.get_json(f"/models/repo/name/{encode_url_param(name)}")

    return build_model_repo(resp, session)


def create_model_repo(
    session: Session,
    name: str,
    description: Optional[str] = None,
    engine: Optional[Engine] = None,
    exists_ok: bool = False,
) -> ModelRepo:
    if not session.is_plan_expired():
        try:
            session.post_json(
                "/models/repo",
                {
                    "modelName": name,
                    "description": description,
                    "modelType": "ludwig",
                    "retrainCadence": "",
                    "retrainConfig": {},
                    "parentID": None,
                    "selectedEngineID": engine.id if engine else None,
                },
            )
            log_info(f"Created model repository: <{name}>")
        except ServerResponseError as e:
            if e.code == HTTPStatus.CONFLICT and exists_ok:
                log_info(f"Model repository {name} already exists and new models will be added to it.")
            else:
                raise e

        # TODO(travis): would be more efficient to return model repo data in CREATE response.
        return get_model_repo(session, name)
    else:
        raise PermissionError(
            "Creating Model Repos is locked for expired plans. Contact us to upgrade or if you would like a demo",
        )



# pylint: disable=W0611,W0613
from typing import Any

import pandas as pd
from IPython.core.magic import cell_magic, Magics, magics_class, needs_local_scope

from predibase import pql


@magics_class
class PQLMagics(Magics):
    def __init__(self, shell):
        super().__init__(shell)

    @needs_local_scope
    @cell_magic("pql")
    def pql(self, line: str, cell: str, local_ns: Any = None) -> pd.DataFrame:
        return pql.execute(cell, params=local_ns)


def _setup_pql_notebook(ipython: Any) -> None:
    ipython.register_magics(PQLMagics)



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

import pandas as pd
from dataclasses_json import config, dataclass_json, LetterCase

# from predibase.connection import Connection
# from predibase.model import Model
# import predibase
from predibase.pql.api import Session
from predibase.resource.connection import Connection

# if TYPE_CHECKING:  # Only imports the below statements during type checking
#     from predibase.resource.connection import Connection
#     from predibase.resource.model import Model
from predibase.resource.dataset_info import DatasetInfoWrapper
from predibase.resource.user import User


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DataField:
    session: Session = field(metadata=config(exclude=True))
    id: int
    name: str
    dataset_id: int = field(metadata=config(field_name="datasetID"))
    _external_connection_id: Optional[int] = field(metadata=config(field_name="extConnectionID"), default=None)
    _external_connection: Optional[Connection] = field(metadata=config(field_name="extConnection"), default=None)

    def set_external_connection(self, ext_connection_id: int):
        self.session.post_json("/data/fields/connections/create", {"id": self.id, "extConnectionID": ext_connection_id})
        self._external_connection_id = ext_connection_id

    def remove_external_connection(self):
        if not self._external_connection_id:
            return

        self.session.post_json("/data/fields/connections/delete", {"id": self.id})
        self._external_connection_id = None
        self._external_connection = None

    @property
    def external_connection(self) -> Optional[Connection]:
        if not self._external_connection_id:
            return None

        if self._external_connection is not None:
            return self._external_connection

        endpoint = f"/connections/{self._external_connection_id}"
        resp = self.session.get_json(endpoint)
        ext_conn = Connection.from_dict({"session": self.session, **resp})
        self._external_connection = ext_conn
        return self._external_connection


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DatasetNaming:
    object_name: str = field(metadata=config(field_name="objectName"))
    dataset_name: str = field(metadata=config(field_name="datasetName"))


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Dataset:
    session: Session = field(metadata=config(exclude=True))
    _id: str = field(metadata=config(field_name="id"))
    _uuid: str = field(metadata=config(field_name="uuid"))
    _name: str = field(metadata=config(field_name="name"))
    _object_name: str = field(metadata=config(field_name="objectName"))
    _created: str = field(metadata=config(field_name="created"))
    _updated: str = field(metadata=config(field_name="updated"))
    _connection_id: int = field(metadata=config(field_name="connectionID"))
    _dataset_info: Optional[DatasetInfoWrapper] = field(metadata=config(field_name="datasetInfo"), default=None)
    _dataset_profile: Optional[Dict] = field(metadata=config(field_name="datasetProfile"), default=None)
    _connection: Optional[Connection] = field(metadata=config(field_name="connection"), default=None)
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)

    def __repr__(self):
        return (
            f"Dataset(id={self.id}, name={self.name}, object_name={self.object_name}, "
            f"connection_id={self.connection_id}, author={self.author}, "
            f"created={self.created}, updated={self.updated})"
        )

    def _build_model_with_backref(self, resp):
        from predibase.resource_util import build_model

        resp["dataset"] = self
        return build_model(resp, self.session)

    @property
    def id(self):
        return self._id

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def row_count(self):
        if self.dataset_info:
            return self.dataset_info.info.row_count
        return None

    @property
    def size_bytes(self):
        if self.dataset_info:
            return self.dataset_info.info.size_bytes
        return None

    @property
    def author(self) -> str:
        return self._user.username if self._user else None

    @property
    def object_name(self):
        return self._object_name

    @property
    def connection_id(self):
        return self._connection_id

    @property
    def connection(self) -> Connection:
        if self._connection is not None:
            return self._connection
        connection = Connection.from_dict({"session": self.session, **self._get()["connection"]})
        self._connection = connection
        return connection

    @property
    def status(self) -> str:
        return self._get()["status"]

    @property
    def dataset_info(self) -> Optional[DatasetInfoWrapper]:
        if self._dataset_info is None:
            resp = self._get(with_dataset_info=True)
            if "datasetInfo" in resp:
                self._dataset_info = DatasetInfoWrapper.from_dict(resp["datasetInfo"])
        return self._dataset_info

    @property
    def dataset_profile(self):
        if self._dataset_profile is None:
            resp = self._get(with_dataset_info=True)
            if "datasetInfo" in resp and "datasetProfile" in resp["datasetInfo"]:
                self._dataset_profile = resp["datasetInfo"]["datasetProfile"]["DatasetProfile"]
        return self._dataset_profile

    @property
    def created(self):
        return self._created

    @property
    def updated(self):
        return self._updated

    def _get(self, with_dataset_info=False):
        resp = self.session.get_json(f"/datasets/{self.id}?withInfo={with_dataset_info}")
        return resp

    def list_models(self) -> List[dict]:
        resp = self.session.get_json(f"/datasets/{self.id}/models")
        dataset = resp["dataset"]
        return [
            self._build_model_with_backref(x["activeModel"]["model"]) for x in dataset["fields"] if "activeModel" in x
        ]

    def get_fields(self, df: bool = False) -> Union[pd.DataFrame, List[DataField]]:
        resp = self.session.get_json(f"/datasets/{self.id}/schema")

        fields = resp["fields"]
        field_data = {"Name": [], "Field_ID": []}

        for f in fields:
            field_data["Name"].append(f["name"])
            field_data["Field_ID"].append(f["id"])

        if df:
            return pd.DataFrame(field_data)

        return [DataField.from_dict({"session": self.session, **x}) for x in resp["fields"]]

    def to_dataframe(self, limit: int = 100000) -> pd.DataFrame:
        """This method converts a Predibase dataset to a pandas DataFrame. By default, it is limited to 100000 rows
        to prevent crashing due to OOM, however, the user can increase the limit if desired.

        Args: limit: Size of the DataFrame to return.

        Returns: pandas.DataFrame
        """
        query = f'SELECT * FROM "{self.name}" LIMIT {limit};'
        return self.session.execute(query, connection_id=self.connection_id)


class DatasetRef:
    def __init__(
        self,
        name: str,
        connection: Union[Connection, str],
        format_properties: Dict[str, str] = None,
    ):
        self.name = name
        self.connection = connection  # can either be a connection object or name
        self.format_properties = format_properties

    def __repr__(self):
        return f"""DatasetRef,
            name: {self.name},
            connection: {self.connection},
            format_properties: {self.format_properties}"""
